package deployment

import (
	"context"
	"fmt"

	"github.com/ankitrgadiya/operatorlib/pkg/interfaces"
	"github.com/ankitrgadiya/operatorlib/pkg/meta"
	"github.com/ankitrgadiya/operatorlib/pkg/pod"
	"github.com/ankitrgadiya/operatorlib/syncer"

	"github.com/pkg/errors"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// GenerateConfigMap generates ConfigMap object as per the `Conf` struct passed
func (c Conf) Mutate() error {
	// var om *metav1.ObjectMeta
	// var container corev1.Container
	//	var initContainer corev1.Container
	//var podTemplate corev1.PodTemplateSpec

	om, err := meta.MutateObjectMeta(c)

	if err != nil {
		return errors.Wrap(err, "failed to generate objectmeta")
	}

	p, err := pod.MutatePod(c.Pod)

	if err != nil {
		return errors.Wrap(err, "failed to generate pod template spec")
	}

	c.Object.ObjectMeta = *om
	c.Object.Spec = appsv1.DeploymentSpec{
		//	Replicas:           c.Size,
		Strategy: c.Strategy,
		Selector: metav1.SetAsLabelSelector(om.Labels),
		Template: corev1.PodTemplateSpec{
			ObjectMeta: *om,
			Spec:       p.Spec,
		},
	}

	return err
}

func (c Conf) MakeObject() *appsv1.Deployment {
	obj := &appsv1.Deployment{
		ObjectMeta: metav1.ObjectMeta{
			Name:      c.GetName(),
			Namespace: c.GetNamespace(),
		},
	}

	return obj
}

func SyncObject(c Conf, ctx context.Context) (syncer.SyncResult, error) {

	c.Object = c.MakeObject()

	PVCSyncer := syncer.NewObjectSyncer("Deployment", c.GetInstance(), c.Object, c.Reconcile, func() error {
		if err := c.Mutate(); err != nil {
			return err
		}
		return nil
	})

	result, err := syncer.Sync(ctx, PVCSyncer)
	if err != nil {
		return result, err
	}
	return result, nil
}

func (c Conf) GetInstance() interfaces.Object {
	return c.Meta.GetInstance()
}
func (c Conf) GetName() string {
	return c.Meta.GetName()
}
func (c Conf) GetNamespace() string {
	return c.Meta.GetNamespace()
}

func (c Conf) GetObject() interfaces.Object {
	return c.Object
}

func (c Conf) GenLabelsFunc() (map[string]string, error) {
	var labels map[string]string
	if c.Meta.GenLabelsFunc != nil {
		return c.Meta.GenLabelsFunc()
	}
	return labels, nil
}
func (c Conf) GenAnnotationsFunc() (map[string]string, error) {
	var annotations map[string]string
	if c.Meta.GenAnnotationsFunc != nil {
		return c.Meta.GenAnnotationsFunc()
	}
	return annotations, nil

}
func (c Conf) GenFinalizersFunc() ([]string, error) {
	if c.Meta.GenFinalizersFunc != nil {
		return c.Meta.GenFinalizersFunc()
	}
	return nil, nil

}

// ObjectType returns the type of the ObjectSyncer subject
func (s Conf) ObjectType() string {
	return fmt.Sprintf("%T", s.GetObject())
}

// OwnerType returns the type of the ObjectSyncer owner
func (s Conf) ObjectOwnerType() string {
	return fmt.Sprintf("%T", s.GetInstance())
}
